<?php
/**
 * The Horde_Auth_Cas class provides transparent authentication
 * through the CAS Web SSO (http://www.jasig.org/cas).  
 *
 * Copyright Consortium Esup-Portail (http://www.esup-portail.org)
 *
 * See the enclosed file COPYING for license information (LGPL). If you did
 * not receive this file, see http://opensource.org/licenses/lgpl-2.1.php
 *
 * @author   Julien Marchal <julien.marchal@univ-nancy2.fr>
 * @author   Xavier Montagutelli <xavier.montagutelli@unilim.fr>
 * @category Horde
 * @license  http://opensource.org/licenses/lgpl-2.1.php LGPL
 * @package  Auth
 */

/* Load the phpCAS module */
include_once('CAS.php');

class Horde_Auth_CAS extends Horde_Auth_Base
{
    /**
     * An array of capabilities, so that the driver can report which
     * operations it supports and which it doesn't.
     *
     * @var array
     */
    protected $_capabilities = array(
        'transparent' => true,
        'list'        => true,
    );

    /**
     * LDAP object, to fetch the users' list from a LDAP server
     *
     * @var Horde_Ldap
     */
    protected $_ldap;

    protected $_session;
    protected $_phpcas_session;

    /** 
     * Name of the imap service (something like imap://imap.corp.com), to retrieve a PT from CAS
     *
     * @var string
     */
    protected $_imapService = "";

    /**
     * ID for Horde caching system for the users' list
     *
     * @connst string
     */
    const cache_id_user_list = 'cas_ldap_userlist';  

    /**
     * Constructor.
     *
     * @param array $params  Parameters:
     * <pre>
     * 'password_header' - (string) Name of the header holding the password of
     *                     the logged in user.
     * 'password_holder' - (string) Where the hordeauth password is stored.
     * 'password_preference' - (string) Name of the Horde preference holding
     *                         the password of the logged in user.
     * 'username_header' - (string) [REQUIRED] Name of the header holding the
     *                     username of the logged in user.
     * </pre>
     *
     * @throws InvalidArgumentException
     */
    public function __construct(array $params = array())
    {
        foreach (array('cas_host') as $val) {
            if (!isset($params[$val])) {
                throw new InvalidArgumentException(__CLASS__ . ': Missing ' . $val . ' parameter.');
            }
        }

        // Default values are merged with configuration supplied values
        $params = array_merge(array(
            'debuglevel' => 'DEBUG',            // debug messages will be emitted with that loglevel
            'cas_host' => '',       
            'cas_port' => '443',
            'cas_context' => '/cas',
            'cas_proxy' => false,               // acts as a CAS proxy ?
            'cas_list_user_cache' => false,     // use Horde caching system to store the users' list ?
            'cas_cacert' => '',
            'cas_pgtdir' => session_save_path(),
            'cas_callback_url' => '',
            'cas_handle_logout' => true,
            'cas_handle_logout_hosts' => array(),
            'uid' => 'uid',
        ), $params);

        if (! empty($params['ldap'])) {
            $this->_debug('Registering the ldap object');
            $this->_ldap = $params['ldap'];
            unset($params['ldap']);
        }

        parent::__construct($params);
        $this->_debug('Building new object, URI ' . $_SERVER['REQUEST_URI']);

        // Restore possible previous arguments
        $this->__get_arg();
   
        // Initialize phpCAS library 
        // For the CAS global logout : the PHP session ID should be named with the ticket ID
        if(!empty($_GET['ticket'])) {  
            $this->_rename_session(preg_replace('/[^\w]/', '', $_GET['ticket']));
        }
        $this->_session = session_id(); // register the PHP session_id, as it appears to change when initializing kronolith. Bug ?

        if (! empty($this->_params['cas_debug'])) {
            phpCAS::setDebug($this->_params['cas_debug']);
        }

        if ($this->_params['cas_proxy']) {
            $this->_debug('CAS Proxy mode');
            if (empty($this->_params['cas_callback_url']))
                throw new InvalidArgumentException(__CLASS__ . ': Missing cas_callback_url parameter.');
            phpCAS::proxy(CAS_VERSION_2_0, $this->_params['cas_host'], $this->_params['cas_port'], $this->_params['cas_context'], false);
            phpCAS::setFixedCallbackURL ($this->_params['cas_callback_url']);
   //         phpCAS::setPGTStorageFile ($this->_params['cas_pgtdir']);
            phpCAS::setPGTStorageDb( $this->_params['cas_db_dsn'],
			$this->_params['cas_db_user'],
		  	$this->_params['cas_db_password'],
		  	$this->_params['cas_db_table'],
		  	null	 
		); 	 

 // JCD : autoriser le passage par un proxy
 phpCAS::setFixedCallbackURL('https://www.u-picardie.fr/horde/casProxy.php');
 phpCAS::setFixedServiceURL('https://www.u-picardie.fr/horde');
 // JCD : autoriser le passage par un proxy

            // Remove trailing "/" for the service URL. Sometimes, the CAS ticket is not validated
            // because the URL does not match the service. They differ with this trailing slash... WHY ?
            $my_url= phpCAS::getServiceURL();
            if (preg_match('&/$&', $my_url)) {
                $my_url = preg_replace('&/$&','', $my_url);
                phpCAS::setFixedServiceURL($my_url);
            }
        } else {
            $this->_debug('CAS Client mode');
            phpCAS::client(CAS_VERSION_2_0, $this->_params['cas_host'], $this->_params['cas_port'], $this->_params['cas_context'], false);
        }

        if (! empty($this->_params['cas_cacert'])) {
            $this->_debug('Verifying cas server certificate with the CA file ' . $this->_params['cas_cacert']);
            phpCAS::setCasServerCACert ($this->_params['cas_cacert']);
        } else {
            $this->_debug('Not verifying the cas server certificate. Not recommended!');
            phpCAS::setNoCasServerValidation();
        }

        // Handle global logout sent by the CAS server
        // "phpCAS can't handle logout requests if it does not manage the session." so we have to register a callback function
        phpCAS::setSingleSignoutCallback (array($this, 'globalLogout'));
        phpCAS::handleLogoutRequests($this->_params['cas_handle_logout']);
        // phpCAS::handleLogoutRequests(
        //         $this->_params['cas_handle_logout'], 
        //         empty($this->_params['cas_handle_logout_host']) ?  $this->_params['cas_host'] : $this->_params['cas_handle_logout_host']
        // );

        $vars = Horde_Variables::getDefaultVariables();
        $this->_logout = $vars->logout_reason;
        $this->_debug('_logout=' . $this->_logout);

        // Hack - Rewrite some PEAR methods to renew password (i.e. proxy ticket) when authentication fails
        // Switchable (on/off) by configuration
        // Only needed when in proxy mode
        if ($this->_params['cas_proxy'] && $this->_params['cas_renew_password']) {
            $this->_debug('Incorporating wrapper functions to renew the Proxy Ticket when IMAP or Sieve login fails (hack).');
            if (! function_exists("runkit_method_rename")) 
                throw new InvalidArgumentException(__CLASS__ . ': parameter "cas_renew_password" is TRUE, but the PHP runkit extension needed by this hack is not available!');
            $this->_wrapImapLoginMethod();
            $this->_wrapTimsievedLoginMethod();
        }

        // HACK : when kronolith is initialized after connecting to Horde, the php session changes and we lose
        // the phpCAS part of the session. We register it to be able to restore it.
        $this->_phpcas_session = $_SESSION['phpCAS'];

        $this->_debug('Done, new object instantiated');
    }

    /**
     * Authentication stub.
     *
     * On failure, Horde_Auth_Exception should pass a message string (if any)
     * in the message field, and the Horde_Auth::REASON_* constant in the code
     * field (defaults to Horde_Auth::REASON_MESSAGE).
     *
     * @param string $userID      The userID to check.
     * @param array $credentials  An array of login credentials.
     *
     * @throws Horde_Auth_Exception
     */
    protected function _authenticate($userId, $credentials)
    {
        throw new Horde_Auth_Exception('Unsupported.');
    }

    /**
     * Automatic authentication: through the CAS server.
     *
     * @return boolean  Whether transparent login is supported.
     * @throws Horde_Auth_Exception
     *
     */
    public function transparent()
    {
        $this->_debug('Starting');
        if (! phpCAS::isAuthenticated()) {
            if ($this->_session !== session_id()) {
                $this->_debug('NOT authenticated AND PHP session mismatch! Trying to restore phpCAS session');
                $_SESSION['phpCAS'] = $this->_phpcas_session;
                $this->_rename_session($this->_session);
                return $this->transparent();
            }
            $this->_debug('NOT authenticated, forcing CAS authentication');
            phpCAS::forceAuthentication();
        }
        $username = phpCAS::getUser();
        if (empty($username)) {
            $this->_log_error('We are supposed to be authenticated by CAS, but no username returned!');
            throw new Horde_Auth_Exception('No username returned by CAS!');
        }
        $this->_debug('Authenticated as user ' . $username);

        // Remove scope from username, if present.
        $this->setCredential('userId', $this->_removeScope($username));

        // Set password for hordeauth login. Only when in proxy mode.
        if ($this->_params['cas_proxy']) {
            $this->_debug('Getting a proxy ticket as the Horde password');
            if (! $this->_getNewPassword()) { 
                throw new Horde_Auth_Exception('Error getting Proxy Ticket!');
            }
        }
        return true;
    }

    /**
     * Checks for triggers that may invalidate the current auth.
     * These triggers are independent of the credentials.
     * Verifie si l'authentification CAS est toujours valide 
     *
     * @return boolean  True if the results of authenticate() are still valid.
     */
    public function validateAuth()
    {
        return phpCAS::isAuthenticated();
    }


    /**
     * List all users in the system. Only method available now : based on a LDAP directory.
     *
     * @return array  The array of userIds.
     * @throws Horde_Auth_Exception
     */
    public function listUsers($sort = false)
    {
        $this->_debug('Listing all users from LDAP');

        if ($this->_params['cas_list_user'] !== 'ldap') return array();

        // First, search inside the cache
        if ($this->_params['cas_list_user_cache']) {
            $this->_debug('Trying to use cached data');
            $cache = $GLOBALS['injector']->getInstance('Horde_Cache');
            $userlist = $cache->get($this::cache_id_user_list, 86400);
            if (! empty($userlist)) return unserialize($userlist);
        } else {
            $this->_debug('Not using the cache');
        }

        // I didn't manage to get automatically a Horde_Ldap object ... TODO
        // The Ldap auth handler seems to receive this object. WHY ?
        if (empty($this->_ldap)) {
            if ($this->_params['driverconfig'] == "horde") {
                foreach ($GLOBALS['conf']['ldap'] as $key => $val) {
                    $this->_params[$key] = $val;
                }
            }
            $this->_ldap = new Horde_Ldap($this->_params);
        }

        // If the users' list is not in the cache, we are doing a LDAP search
        $this->_debug('Doing a LDAP lookup with object ' . get_class($this->_ldap));
        $params = array(
            'attributes' => array($this->_params['uid']),
            'scope' => $this->_params['scope'],
            'sizelimit' => isset($this->_params['sizelimit']) ? $this->_params['sizelimit'] : 0
        );
        $userlist = array();
        try {
            $search = $this->_ldap->search(
                $this->_params['basedn'],
                $this->_params['filter'],
                $params);
            $uid = Horde_String::lower($this->_params['uid']);
            while ($entry = $search->shiftEntry()) {
                $userlist[] = $entry->getValue($uid, 'single');
            }
        } catch (Horde_Ldap_Exception $e) {
            $this->_log_error('LDAP search error : ' . $e);
            throw $e;
        }

        $userlist = $this->_sort($userlist, $sort);
        // Record the result in the cache, if needed
        if ($this->_params['cas_list_user_cache']) {
            $cache->set($this::cache_id_user_list, serialize($userlist), 86400);
        }
        return $userlist;
    }

    /**
     * Callback function called when the CAS server sends a "global logout" request
     * This function is not called by the user or its browser, but from the CAS server!
     *
     * The CAS POSTs a SAML message containing the ST initially delivered to Horde for a user disconnecting from CAS
     * We have previously renamed the user session id with the ST => here, we create a new PHP session with that
     * "secret" and we destroy it. Code stolen from the phpCAS library.
     *
     * @param string $ticket2logout CAS Service Ticket linked to a user that is disconnecting from CAS server
     *
     */
    function globalLogout($ticket2logout)
    {
        $this->_debug('Global Logout called for ticket ' . $ticket2logout);

        $session_id = preg_replace('/[^\w]/','',$ticket2logout);

        // destroy a possible application session created before phpcas
        if(session_id() !== ""){
            session_unset();
            session_destroy();
        }
        // fix session ID
        session_id($session_id);
        $_COOKIE[session_name()]=$session_id;
        $_GET[session_name()]=$session_id;

        // Overwrite session
        session_start();
        session_unset();
        session_destroy();
    }

    /**
     * -=-=-=- Private methods -=-=-=-=-
     */

    /**
     * Log messages for debugging
     *
     */
    protected function _debug($message)
    {
        if (! isset($this->_logger)) return;
        $call = debug_backtrace(false);
        $call = (isset($call[1]))?$call[1]:$call[0];
        $header = __CLASS__ . '::' . $call['function'] . ' ' . $_SERVER['REMOTE_ADDR'] . ' [' . session_id() . '][' . $this->_session . ']';
        $this->_logger->log($header . $message, $this->_params['debuglevel']);
    }
    protected function _log_error($message)
    {
        if (! isset($this->_logger)) return;
        $call = debug_backtrace(false);
        $call = (isset($call[1]))?$call[1]:$call[0];
        $header = __CLASS__ . '::' . $call['function'] . ' ';
        $this->_logger->log($header . $message, 'ERR');
    }

    /**
     * Rename the PHP session
     *
     * @param string $id New PHP session ID
     */
    protected function _rename_session($id) 
    {
        $old_session = $_SESSION;
        session_destroy();
        session_id($id);
        session_start();
        $_SESSION = $old_session;
    }
     
    /**
     * Removes the scope from the user name, if present.
     *
     * @param string $username  The full user name.
     *
     * @return string  The user name without scope.
     */
   protected function _removeScope($username)
   {
       $pos = strrpos($username, '@');
       return ($pos !== false)
           ? substr($username, 0, $pos)
           : $username;
   }

   /**
    * Rewrite dynamically some PEAR methods ...
    * Hack (Bad Thing) to renew the PT (password) in case of authentication failure
    * inside a Horde App like IMP when connecting an IMAP server.
    */
    protected function _wrapImapLoginMethod()
    {
        $filename = HORDE_BASE . '/lib/Cas/RunkitRedefine_Horde_Imap_Client_Socket::_tryLogin.php';
        $imapLogin = file_get_contents($filename);
        if (! $imapLogin || empty ($imapLogin)) {
            $this->_log_error('Error reading file containing wrapper Horde_Imap_Client_Socket::_tryLogin method!');
            throw new InvalidArgumentException(__CLASS__ . ': parameter "cas_renew_password" is TRUE, but the wrapper function for the Imap Login function is not available!');
            return false;
        }

        include_once('Horde/Imap/Client/Socket.php'); // To load the class and the method to redefine
        if (! runkit_method_rename('Horde_Imap_Client_Socket', '_tryLogin', '__tryLogin')) {
            $this->_log_error('Error renaming Horde_Imap_Client_Socket::_tryLogin method!');
            return false;
        }
        if (! runkit_method_add('Horde_Imap_Client_Socket', '_tryLogin', '$method', $imapLogin)) {
            $this->_log_error('Error adding wrapper Horde_Imap_Client_Socket::_tryLogin method!');
            return false;
        }
        $this->_debug('Wrapper for method Horde_Imap_Client_Socket::_tryLogin loaded');
        return true;
    }

    protected function _wrapTimsievedLoginMethod()
    {
        $filename = HORDE_BASE . '/lib/Cas/RunkitRedefine_Ingo_Transport_Timsieved::_connect.php';
        $timsievedConnect = file_get_contents($filename);
        if (! $timsievedConnect || empty ($timsievedConnect)) {
            $this->_log_error('Error reading file containing wrapper Ingo_Transport_Timsieved::_connect method!');
            throw new InvalidArgumentException(__CLASS__ . ': parameter "cas_renew_password" is TRUE, but the wrapper function for the Seieve Login function is not available!');
            return false;
        }

        include_once(HORDE_BASE . '/ingo/lib/Transport.php'); // To load the class and the method to redefine
        include_once(HORDE_BASE . '/ingo/lib/Transport/Timsieved.php'); // To load the class and the method to redefine
        if (! runkit_method_rename('Ingo_Transport_Timsieved', '_connect', '__connect')) {
            $this->_log_error('Error renaming Ingo_Transport_Timsieved::_connect method!');
            return false;
        }
        if (! runkit_method_add('Ingo_Transport_Timsieved', '_connect', '', $timsievedConnect)) {
            $this->_log_error('Error adding wrapper Ingo_Transport_Timsieved::_connect method!');
            return false;
        }
        $this->_debug('Wrapper for method Ingo_Transport_Timsieved::_connect loaded');
        return true;
    }

    /**
     * Get a new Proxy Ticket, stored as the password for Horde (and other apps)
     *
     * @access public
     *
     * @return string  The new password, or false
     */
    protected function _getNewPassword($service='')
    {
        if (! phpCAS::isAuthenticated()) {
            $this->_log_error('Trying to renew CAS Proxy Ticket, but phpCAS says we are not authenticated!');
            return false;
            // $this->_refreshAuth();
        }

        if (empty ($service)) {
            if (empty($this->_imapService)) $this->_getImapService();
            $service = $this->_imapService;
        }
        $pt = $this->_getPT($service);
        $this->_debug('New Proxy Ticket ' . $pt . ' for user ' . phpCAS::getUser());
        $this->setCredential('credentials', array( 'password' => $pt));
        //?? Auth::clearAuth();
        return $pt;
    }

    protected function _getPT($service, $refresh=false)
    {
        $pt = phpCAS::retrievePT($service, $err_code, $output);
        if (empty($pt) || 
            (!preg_match('/^PT-/',$pt) && !preg_match('/^ST-/',$pt)) ) {
            $this->_log_error('User ' . phpCAS::getUser() . " Bad or empty Proxy Ticket ($pt) ! phpCAS error : " . $output);
            $this->_refreshAuth();
            throw new Horde_Auth_Exception("Bad or empty Proxy Ticket. You must login again ?");
        } 
        $this->_debug('User ' . phpCAS::getUser() . " retrieve Proxy Ticket $pt");
        return($pt);
    }

    protected function _refreshAuth() {
        // Hack to force phpCAS renewal ...
        $this->_debug('Starting');
        unset ($_SESSION['phpCAS']);
        $GLOBALS['registry']->clearAuth();
        $this->__sav_arg();
        phpCAS::renewAuthentication();
    }

    /**
     *
     */
    protected function _getImapService()
    {
        $servers = Horde::loadConfiguration('backends.php', 'servers', 'imp');
        @define ('IMP_BASE', HORDE_BASE . '/imp');
        require_once IMP_BASE . '/lib/Auth.php';
        require_once IMP_BASE . '/lib/Imap.php';
        require_once IMP_BASE . '/lib/Imap/Config.php';

        $server = IMP_Auth::getAutoLoginServer();

        if (empty($server)) return false;

        $p = preg_replace("'/(.*)$'" , "" , $servers[$server]['protocol']);
        $host = empty($servers[$server]['realhost']) ? $servers[$server]['hostspec'] : $servers[$server]['realhost'];
        $this->_imapService = $p . '://' . $host;
    	$this->_debug('User ' . phpCAS::getUser() . ', IMAP server key ' . $server . ' (web server name ' . $_SERVER['SERVER_NAME'] . '), IMAP service ' . $this->_imapService);
        return $this->_imapService;
    }

    /**
     * Save the URL arguments in a temporary file
     * For future use after authentication
     */
    protected function __sav_arg() 
    {
        if (empty($_GET['pgtId']) &&
            empty($_SESSION['cas']['__SAVARG']) ) {

            if(!empty($_GET)  && is_array($_GET))  $SAV['_GET']  = $_GET;
            if(!empty($_POST) && is_array($_POST)) $SAV['_POST'] = $_POST;

            $_SESSION['cas']['__SAVARG_FILE'] = $this->__tempFilePath();

            $fp = @fopen($_SESSION['cas']['__SAVARG_FILE'], 'w');
            if(!$fp) {
                $this->_log_error('Failed to open save argument file : ' . $_SESSION['cas']['__SAVARG_FILE']);
                unset($_SESSION['cas']['__SAVARG_FILE']);
                return(false);
            }
            fwrite( $fp, serialize($SAV));
            fclose($fp);
        }
    }

    protected function __get_arg() 
    {
        if (empty($_SESSION['cas']['__SAVARG_FILE']) ||
            ! file_exists($_SESSION['cas']['__SAVARG_FILE'])) {
            unset ($_SESSION['cas']['__SAVARG_FILE']);
            return false;
        }

        $serial_sav = file_get_contents($_SESSION['cas']['__SAVARG_FILE']);
        if (! $serial_sav || empty($serial_sav)) {
                $this->_log_error('Failed to read content of save argument file : ' . $_SESSION['cas']['__SAVARG_FILE']);
                unset($_SESSION['cas']['__SAVARG_FILE']);
                return false;
        }
        $SAV = unserialize($serial_sav);
        @unlink($_SESSION['cas']['__SAVARG_FILE']);

        if(!empty($SAV['_GET']) && is_array($SAV['_GET'])) {
            reset ($SAV['_GET']);
            while (list ($key, $val) = each ($SAV['_GET'])) {
                if(empty($_GET[$key])){
                    if($key == 'url') {
                        $val = preg_replace("/&amp;/","&",$val);
                        $val = Horde::applicationUrl($val,true,false);
                    }
                    $_GET[$key] = $val;
                }
            }
        }

        if(!empty($SAV['_POST']) && is_array($SAV['_POST'])) {
            reset ($SAV['_POST']);
            while (list ($key, $val) = each ($SAV['_POST'])) {
                if(empty($_POST[$key])) {
                    if($key == 'url') {
                        $val = preg_replace("/&amp;/","&",$val);
                        $val = Horde::applicationUrl($val,true,false);
                    }
                    $_POST[$key] = $val;
                }
            }
        }
        $_SESSION['cas']['__SAVARG'] = '';
        unset($_SESSION['cas']['__SAVARG']);
    }

    protected function __tempFilePath() 
    {
        return Horde::getTempFile('cas_', false);
    }

    /**
     * Converts Cas connection parameters to Horde_Ldap parameters (stolen from Turba code).
     *
     * @param array $in  Cas parameters.
     *
     * @return array  Horde_Ldap parameters.
     */
    protected function _convertParameters(array $in)
    {
        $map = array(
            'cas_ldap_host' => 'hostspec',
            'cas_ldap_port' => 'port',
            'cas_ldap_version' => 'version',
            'cas_ldap_basedn' => 'basedn',
            'cas_ldap_binddn' => 'binddn',
            'cas_ldap_bindpw' => 'bindpw',
            // can both be specified in Turba but only one in Horde_Ldap.
            //'objectclass',
            //'filter' => 'filter',
            'cas_ldap_scope' => 'scope',
            // charset is always utf-8
            //'charset',
            // Not yet implemented.
            //'deref',
            //'referrals',
            //'sizelimit',
            //'dn',
        );
        $out = array();
        foreach ($in as $key => $value) {
            if (isset($map[$key])) {
                $out[$map[$key]] = $value;
            }
        }
        return $out;
    }

}

